import unittest
from functools import partial
import math
import json
from artifax import build, At
from artifax.exceptions import UnresolvedDependencyError

class BuildTest(unittest.TestCase):

    def test_empty_build(self):
        artifacts = {}
        result = build(artifacts)
        self.assertEqual(result, {})

    def test_single_artifact_build(self):
        artifacts = {'a': 42}
        result = build(artifacts)
        self.assertEqual(len(result), 1)
        self.assertEqual(result, {'a': 42})

    def test_artifact_immutability(self):
        artifacts = {
            'a': lambda: 42,
            'b': lambda a: a()**2
        }

        results = build(artifacts, solver='async')

        self.assertTrue(isinstance(artifacts, dict))
        self.assertEqual(len(artifacts), 2)
        self.assertEqual(set(artifacts.keys()), {'a', 'b'})
        self.assertTrue(callable(artifacts['b']))
        self.assertFalse(callable(results['b']))

    def test_constant_artifacts_build(self):
        obj = object()
        artifacts = {
            'int': 42,
            'string': 'Hello',
            'obj': obj,
            'list': [1,2,3],
            'dictionary': {'answer': 42},
            'set': set([1,2,3.14])
        }
        result = build(artifacts, solver='linear') # async solver copies objects to subprocess!!
        for k in artifacts:
            self.assertEqual(result[k], artifacts[k])

    def test_sample_build(self):
        artifacts = {
            'A': 42,
            'B': lambda: 7,
            'C': lambda: 10,
            'AB': lambda A, B: A + B(),
            'C minus B': lambda B, C: C() - B(),
            'greet': 'Hello',
            'msg': lambda greet, A: '{} World! The answer is {}.'.format(greet, A),
        }
        result = build(artifacts)
        self.assertEqual(result['AB'], 49)
        self.assertEqual(result['C minus B'], 3)
        self.assertEqual(result['msg'], 'Hello World! The answer is 42.')

    def test_partial_build(self):
        artifacts = {
            'A': lambda x: x**2,
            'B': lambda A, x: A(x),
            'C': lambda A, x: 'A(4)-{} is equal to {}'.format(x, A(4)-x)
        }
        result = build(artifacts, allow_partial_functions=True)
        self.assertTrue(callable(result['A']))
        self.assertEqual(result['A'](4), 16)
        self.assertEqual(result['B'](-5), 25)
        self.assertEqual(result['C'](6), 'A(4)-6 is equal to 10')

    def test_build_with_partial_functions(self):
        artifacts = {
            'a': 42,
            'b': lambda A: A*2
        }
        with self.assertRaises(UnresolvedDependencyError):
            _ = build(artifacts, solver='linear')

        result = build(artifacts, allow_partial_functions=True)
        self.assertIsInstance(result['b'], partial)

    # TODO: fix those!
    # def test_special_keys(self):
    #     artifacts = {
    #         'key-with-dash': 'a value',
    #         'bang': lambda key_with_dash: '{}!'.format(key_with_dash),
    #         '_underscore_key': lambda bang: '_{}_'.format(bang),
    #     }
    #     result = build(artifacts)
    #     self.assertEqual(result['key-with-dash'], 'a value')
    #     self.assertEqual(result['bang'], 'a value!')
    #     self.assertEqual(result['_underscore_key'], '_a value!_')

    def test_at_constructor(self):
        def subtract(p, q):
            return p - q

        result = build({
            'p': 3,
            'q': 5,
            'p - q': subtract,
            'q - p': lambda p, q: subtract(q, p),
        })

        self.assertEqual(result['p - q'], -2)
        self.assertEqual(result['q - p'], 2)

        result = build({
            'a': -11,
            'b': 7.5,
            'a - b': At('a', 'b', subtract),
            'b - a': At('b', 'a', subtract),
        })

        self.assertEqual(result['a - b'], -18.5)
        self.assertEqual(result['b - a'], 18.5)

    def test_deep_build(self):
        for solver in ['linear', 'bfs', 'bfs_parallel', 'async']:
            results = build({
                'a': 'a',
                'b': 'b',
                'x': lambda a: 'x',
                'c': lambda x, b: 'c',
            }, solver=solver)

            self.assertEqual(results['c'], 'c')

    def test_solvers(self):
        def subtract(p, q):
            return p - q

        solvers = ['linear', 'bfs', 'bfs_parallel', 'async']
        results = [build({
            'a': -11,
            'b': 7.5,
            'a - b': At('a', 'b', subtract),
            'b - a': At('b', 'a', subtract),
        }, solver=solver) for solver in solvers]

        # make sure we have as many results as there are solvers
        self.assertEqual(len(results), len(solvers))

        # serialize results and add them to set
        # if results are the same, there should be only one element in the set
        result_set = set([json.dumps(result, sort_keys=True) for result in results])
        self.assertEqual(len(result_set), 1)

        result = json.loads(next(iter(result_set)))
        self.assertEqual(result['a - b'], -18.5)
        self.assertEqual(result['b - a'], 18.5)



"""artifax is a Python package to evaluate nodes in a computation graph where
the dependencies associated with each node are extracted directly from their
function signatures.
"""

from artifax.builder import *
from artifax.exceptions import *
from artifax.utils import *
from artifax.models import *
import artifax.io



""" models.py

This module hosts classes that provided the object-oriented support to building
artifacts.
"""

from functools import reduce
import operator
from exos import each
from . import builder
from . import utils as u


def _fluent(cls, attr, *args):
    """ provides a fluent interface for any classes that choose to apply it. """
    if args:
        setattr(cls, attr, args[0])
        return cls
    return getattr(cls, attr)


class Artifax:
    """ The Artifax class enables artifacts to be built through a conventional
    object-oriented interface. Its stateful nature boasts additional capabilities
    like incremental builds and building of stale nodes only.
    """
    class Result:
        """ The Result class acts as an augmented dictionary to
        hold the artifax build products and any additional information
        deemed necessary or interesting. """
        def __init__(self, *args, **kwargs):
            self._data = {}
            self.update(*args, **kwargs)

        def __setitem__(self, key, item):
            self._data[key] = item

        def __getitem__(self, key):
            return self._data[key]

        def __repr__(self):
            return repr(self._data)

        def __len__(self):
            return len(self._data)

        def __delitem__(self, key):
            del self._data[key]

        def clear(self):
            """ Removes all items from result. """
            return self._data.clear()

        def copy(self):
            """ Returns a shallow copy of the result dictionary. """
            return self._data.copy()

        def has_key(self, k):
            """ Returns True if key k is part of the result and False otherwise.
            """
            return k in self._data

        def update(self, *args, **kwargs):
            """ R.update([E, ]**F) -> None.  Update R from dict/iterable E and F.
            If E is present and has a .keys() method, then does:  for k in E: R[k] = E[k]
            If E is present and lacks a .keys() method, then does:  for k, v in E: R[k] = v
            In either case, this is followed by: for k in F:  R[k] = F[k] """
            return self._data.update(*args, **kwargs)

        def keys(self):
            """ Returns a set-like object providing a view on D's keys. """
            return self._data.keys()

        def values(self):
            """ Returns an object providing a view on D's values. """
            return self._data.values()

        def items(self):
            """ Returns key, value pairs of data dictionary items. """
            return self._data.items()

        def pop(self, *args):
            """ pop(k, [,d]) -> v, remove specified key and return
            the corresponding value. If key is not found, d is returned
            if given, otherwise KeyError is raised """
            return self._data.pop(*args)

        def __cmp__(self, dict_):
            return self._data == dict_

        def __contains__(self, item):
            return item in self._data

        def __iter__(self):
            return iter(self._data)

    def __init__(self, dic=None, allow_partial_functions=False, **kwargs):
        if dic is None:
            dic = {}
        dic.update(kwargs)
        self._artifacts = dic.copy()
        self._graph = None
        self._update_graph()
        self._result = Artifax.Result()
        self._stale = set(self._artifacts.keys())
        self._allow_partial_functions = allow_partial_functions

    def _update_graph(self):
        self._graph = u.to_graph(self._artifacts)

    def set(self, *args, **kwargs):
        """ Sets node value. """
        if kwargs:
            for key, value in kwargs.items():
                self.set(key, value)
            return
        node, value = args[0], args[1]
        if node in self._artifacts:
            self._revoke(node)
        self._stale.add(node)
        self._artifacts[node] = value
        self._update_graph()

    def _revoke(self, node):
        self._stale.add(node)
        each(self._revoke, self._graph[node])

    def pop(self, node):
        """ Removes node from the artifacts. """
        if node in self._stale:
            self._stale.remove(node)
        item = self._artifacts.pop(node)
        self._update_graph()
        return item

    def _shipment(self, targets=None):
        nodes = (
            self._stale if targets is None else
            set(reduce(
                operator.iconcat,
                [self._dependencies(t) for t in targets] + [list(targets)],
                []
            ))
        )
        return {
            k: self._artifacts[k]
            for k in nodes
        }

    def _dependencies(self, node):
        def _moonwalk(node, graph, dependencies):
            for vertex, neighbors in graph.items():
                if node in neighbors:
                    dependencies.append(vertex)
                    _moonwalk(vertex, graph, dependencies)

        dependencies = []
        _moonwalk(node, self._graph, dependencies)
        return dependencies

    def build(self, targets=None, allow_partial_functions=None, solver='linear', **kwargs):
        """ Builds artifacts. Returns either a dictionary of resolved nodes or a tuple
        where each item corresponds to one of the defined targets.

        Args:
            targets (:obj:`string or tuple`, optional): Defines specific targets
                to be built. Either a tuple of node names or a string for single
                targets.
            allow_partial_functions (bool, optional): Set to True if artifacts are
                allowed to be resolved to partial functions. Defaults to False.
            solver (str, optional): Choose artifax solver strategy. Pick between
                {'linear', 'bfs', 'bfs_parallel', 'async'}. Defaults to 'linear'.
                Throws InvalidSolverError if solver is not among the available options.
            **kwargs: Arbitrary keyword arguments that are solver-specific.
        """
        targets = (targets,) if isinstance(targets, str) else targets
        if targets:
            for target in targets:
                if target not in self:
                    raise KeyError(target)

        shipment = self._shipment(targets)
        result = builder.build(
            shipment,
            solver=solver,
            allow_partial_functions=(
                allow_partial_functions if allow_partial_functions is not None else
                self._allow_partial_functions
            ),
            **kwargs
        )

        self._stale = {k for k in self._stale if k not in shipment}
        self._result.update(result)

        if targets is None:
            return self._result

        payload = tuple([self._result[target] for target in targets])
        return payload if len(payload) > 1 else payload[0]

    def initial(self):
        """ Returns the initial objects of the artifacts graph, that is,
        the nodes that have no incoming edges, no dependencies."""
        return u.initial(self._graph)

    def number_of_edges(self):
        """ Returns the number of edges in the artifacts graph."""
        return sum([len(v) for v in self._graph.values()])

    def number_of_nodes(self):
        """ Returns the number of nodes in the artifacts graph."""
        return len(self._artifacts)

    def __len__(self):
        return self.number_of_nodes()

    def __contains__(self, node):
        return node in self._artifacts



from functools import reduce
from jinja2 import Template
import os
from . import utils


def tex(artifacts):
    graph = utils.to_graph(artifacts)
    nodes = utils.topological_sort(graph)

    def _edge_reducer(pairs, key):
        for value in graph[key]:
            pairs.append((key, value))
        return pairs

    with open(os.path.join(os.path.dirname(__file__), 'templates', 'dag.tex.j2')) as handle:
        template = Template(handle.read())

    nodes = list(artifacts.keys())
    edges = reduce(_edge_reducer, graph.keys(), [])
    return template.render(nodes=nodes, edges=edges)



""" builder.py

This module hosts the core build function and the private functions that aid it.
"""

from functools import reduce, partial
from collections import deque
import pathos.multiprocessing as mp
from exos import compose, each
from . import utils as u
from .exceptions import UnresolvedDependencyError, InvalidSolverError

# pylint: disable=C0103
_apply = lambda v, *args: (
    v(*args) if callable(v) and args and len(u.arglist(v)) == len(args) else
    partial(v, *args) if callable(v) else
    v
)


def build(artifacts, allow_partial_functions=False, solver='linear', **kwargs):
    """ Core artifact building function. Given an input dictionary describing the
    computation graph where each vertex correspond to a key and edges can be extracted
    from the function signatures associated with each key, the build function returns
    a new dictionary where each key is mapped to its final value.

    Args:
        allow_partial_functions (bool, optional): Set to True if artifacts are
            allowed to be resolved to partial functions. Defaults to False.
        solver (str, optional): Choose artifax solver strategy. Pick between
            {'linear', 'bfs', 'bfs_parallel', 'async'}. Defaults to 'linear'.
        **kwargs: solver-specific keyword arguments.
    """
    solvers = {
        'linear': _build_linear,
        'bfs': _build_bfs,
        'bfs_parallel': _build_parallel_bfs,
        'async': _build_async
    }
    if solver not in solvers:
        raise InvalidSolverError('unrecognized solver [{}]'.format(solver))

    return solvers[solver](
        artifacts.copy(),
        apf=allow_partial_functions,
        **kwargs
    )


def _build_linear(artifacts, apf=False):

    def _reducer(store, node):
        store[node] = _resolve(node, store, apf=apf)
        return store

    nodes = compose(u.topological_sort, u.to_graph)(artifacts)
    return reduce(_reducer, nodes, artifacts)


def _pendencies(graph, node, done):
    return {
        k for k, v in graph.items() if node in v and k not in done
    }


def _build_bfs(artifacts, apf=False):
    done = set()
    graph = u.to_graph(artifacts)
    frontier = deque(u.initial(graph))
    while frontier:
        node = frontier.popleft()
        artifacts[node] = _resolve(node, artifacts, apf=apf)
        done.add(node)
        frontier += [nxt for nxt in graph[node] if not _pendencies(graph, nxt, done)]

    return artifacts


def _build_parallel_bfs(artifacts, apf=False, processes=None):
    done = set()
    graph = u.to_graph(artifacts)
    frontier = set(u.initial(graph))
    pool = mp.Pool(processes=processes)
    while frontier:
        artifacts.update({
            node: pool.apply(_resolve, args=(node, artifacts))
            for node in frontier
        } if len(frontier) > 1 else {
            node: _resolve(node, artifacts, apf=apf)
            for node in frontier
        })
        done |= frontier
        frontier = reduce(lambda acc, curr: acc | curr, [
            {nxt for nxt in graph[node] if not _pendencies(graph, nxt, done)}
            for node in frontier
        ], set())

    pool.close()

    return artifacts


def _build_async(artifacts, apf=False, processes=None):
    graph = u.to_graph(artifacts)
    frontier = u.initial(graph)

    if not frontier:
        return {}

    done, rem = set(), set()
    if processes is None:
        processes = min(max(1, mp.cpu_count() - 1), len(frontier))
    pool = mp.Pool(processes=processes)

    each(lambda node: pool.apply_async(
        partial(_resolve, apf=apf),
        args=(node, artifacts),
        callback=partial(_on_done, artifacts, graph, node, done, rem, apf=apf)
    ), frontier)

    pool.close()
    pool.join()

    artifacts.update({
        n: _resolve(n, artifacts, apf=apf)
        for n in rem
    })

    return artifacts


def _on_done(artifacts, graph, node, done, rem, value, apf=False, **kwargs):
    done.add(node)
    artifacts[node] = value

    frontier = rem | set(graph[node])
    batch = {n for n in frontier if not _pendencies(graph, n, done)}
    rem = frontier - batch

    if not batch:
        return

    pool = mp.Pool(processes=min(max(1, mp.cpu_count() - 1), len(batch)))

    each(lambda node: pool.apply_async(
        _resolve,
        args=(node, artifacts),
        callback=partial(_on_done, artifacts, graph, node, done, rem, apf=apf, **kwargs)
    ), batch)

    pool.close()
    pool.join()


def _resolve(node, store, apf=False):
    value = store[node]
    args = u.arglist(value)
    if isinstance(value, u.At):
        args = value.args()
        value = value.value()
    keys = [u.unescape(a) for a in args]
    args = [store[key] for key in keys if key in store]
    unresolved = [key for key in keys if key not in store]
    if not apf and unresolved:
        raise UnresolvedDependencyError("Cannot resolve {}".format(unresolved))
    return _apply(value, *args)



class CircularDependencyError(Exception):
    """ Exception to be thrown when artifacts can not be built due to the fact
    that there is at least one closed loop in its graph representation which
    means we can not determine an evaluation order for the artifact nodes"""


class UnresolvedDependencyError(Exception):
    """ This exception is thrown when not all of a node's depencies can be found
    in the artifax graph. If you do want any of your nodes to resolve to a partial
    function, you need to set the allow_partial_functions flag in the build
    method/function to True."""


class InvalidSolverError(Exception):
    """ Thrown when requested solver is not available. """




from setuptools import setup
from os import path

with open('requirements.txt') as handle:
    requirements = handle.readlines()

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='artifax',
    version='0.1.3.1',
    packages=['artifax'],
    description='python package for building artifacts from a computational graph',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Bruno Lange',
    author_email='blangeram@gmail.com',
    url='https://gitlab.com/brunolange/artifax',
    install_requires=requirements,
    python_requires='>=3',
    extras_require={
        'dev': [
            'pylint'
        ]
    }
)



import unittest
from artifax import io

class BuildTest(unittest.TestCase):

    def test_sample(self):
        artifacts = {
            'A': 42,
            'B': lambda: 7,
            'C': lambda: 10,
            'AB': lambda A, B: A + B(),
            'C minus B': lambda B, C: C() - B(),
            'greeting': 'Hello',
            'message': lambda greeting, A: '{} World! The answer is {}.'.format(greeting, A),
        }
        lines = io.tex(artifacts).split('\n')

        expected_nodes = [
            '\\node[state] ({}) {{${}$}};'.format(node, node)
            for node in artifacts
        ]

        expected_edges = [
            '(A) edge [mystyle] (AB)',
            '(B) edge [mystyle] (AB)',
            '(B) edge [mystyle] (C minus B)',
            '(C) edge [mystyle] (C minus B)',
            '(greeting) edge [mystyle] (message)'
        ]

        report = lambda l: 'missing line: [{}]'.format(l)
        for expected_line in expected_nodes + expected_edges:
            self.assertTrue(expected_line in lines, report(expected_line))



import unittest
import math
from functools import partial
from artifax import Artifax
from artifax.exceptions import UnresolvedDependencyError
from artifax.utils import At

class ModelTest(unittest.TestCase):

    def test_add(self):
        afx = Artifax()
        afx.set('a', 42)
        self.assertEqual(len(afx), 1)
        self.assertEqual(afx.number_of_nodes(), 1)
        self.assertEqual(afx.number_of_edges(), 0)
        self.assertTrue('a' in afx)

        afx.set('x', lambda a: a/3.14)
        self.assertEqual(len(afx), 2)
        self.assertEqual(afx.number_of_nodes(), 2)
        self.assertEqual(afx.number_of_edges(), 1)
        self.assertTrue('x' in afx)

        self.assertFalse('y' in afx)

    def test_pop(self):
        afx = Artifax()
        afx.set('c', 'C')
        c = afx.pop('c')
        self.assertTrue(len(afx) == 0)
        self.assertEqual(c, 'C')

    def test_build(self):
        obj = object()
        afx = Artifax({
            'int': 42,
            'float': 1.618,
            'string': 'Hello',
            'obj': obj,
            'list': [1,2,3],
            'dictionary': {'answer': 42},
            'set': set([1,2,3.14])
        })
        result = afx.build(solver='linear')
        self.assertEqual(result['int'], 42)
        self.assertEqual(result['float'], 1.618)
        self.assertEqual(result['string'], 'Hello')
        self.assertEqual(result['obj'], obj)
        self.assertEqual(result['list'], [1,2,3])
        self.assertEqual(result['dictionary'], {'answer': 42})
        self.assertEqual(result['set'], set([1,2,3.14]))

    def test_invalid_build(self):
        artifacts = {
            'a': 42,
            'b': lambda A: A*2
        }
        afx = Artifax(artifacts)
        with self.assertRaises(UnresolvedDependencyError):
            _ = afx.build(solver='bfs_parallel')

        result = afx.build(allow_partial_functions=True)
        self.assertIsInstance(result['b'], partial)

    def test_incremental_build(self):
        class ExpensiveObject:
            def __init__(self):
                self.counter = 0
            def expensive_method(self, _):
                self.counter += 1
                return 'foobar'

        exo = ExpensiveObject()
        afx = Artifax(
            p=(3,4),
            q=(12, 13),
            exo=lambda q: exo.expensive_method(q),
        )

        # pool_async silently fails to get trigger callback that resolves
        # the nodes, I guess because it can't pickle ExpensiveObject
        # from the unittest thread
        result = afx.build(solver='linear')
        self.assertEqual(exo.counter, 1)

        afx.set('p', (1,1))
        result = afx.build(solver='linear')
        self.assertEqual(exo.counter, 1)

        afx.set('q', (0,0))
        result = afx.build(solver='linear')
        self.assertEqual(exo.counter, 2)

        afx.set('new', 'hello')
        result = afx.build(solver='linear')
        self.assertEqual(result['new'], 'hello')
        self.assertEqual(exo.counter, 2)

        afx.pop('new')
        result = afx.build(solver='linear')
        self.assertEqual(exo.counter, 2)

    def test_initial(self):
        afx = Artifax({
            'earth': object(),
            'un': lambda earth: 'water @{}'.format(earth),
            'mars': object(),
            'mcrn': lambda mars: 'dust @{}'.format(mars),
            'belt': object(),
            'opa': lambda belt: 'ice @{}'.format(belt)
        })
        self.assertEqual(afx.initial(), set(['earth', 'mars', 'belt']))
        _ = afx.build()

    def test_multiprocessing(self):
        afx = Artifax({
            'p1': 'p1',
            'p2': 'p2',
            'c1': lambda p1: 'c1 after {}'.format(p1),
            'c2': lambda p1: None,
            'c3': lambda p2: None,
            'c4': lambda p2: None,
            'c5': lambda p1, p2, c2, c3: None,
            'c6': lambda c1: None,
            'F':  lambda c1, c2, c3, c4, c5: None,
        })

        c1 = afx.build(targets=('c1',))
        self.assertEqual(c1, 'c1 after p1')

    def test_in_operator(self):
        afx = Artifax(p=(3,4))
        self.assertTrue('p' in afx)
        self.assertFalse('q' in afx)

        afx.pop('p')
        self.assertFalse('p' in afx)

        afx.set('q', (1,2))
        self.assertTrue('q' in afx)

    def test_targeted_build(self):
        afx = Artifax(
            {'greeting': lambda name, punctuation: 'Hello, {}{}'.format(name, punctuation)},
            name='World',
            punctuation=''
        )
        greeting = afx.build(targets='greeting')
        self.assertEqual(greeting, 'Hello, World')

        afx.set('punctuation', '!')
        greeting, punctuation = afx.build(targets=('greeting', 'punctuation'))
        self.assertEqual(greeting, 'Hello, World!')
        self.assertEqual(punctuation, '!')

        result = afx.build()
        self.assertEqual(result['greeting'], 'Hello, World!')
        self.assertEqual(result['name'], 'World')
        self.assertEqual(result['punctuation'], '!')

    def test_stale(self):
        class C:
            counter = 0
            def __init__(self):
                C.counter += 1

        afx = Artifax({
            'a': 42,
            'b': lambda a: math.pow(a, 5),
            'counter': lambda: C(),
        })
        result = afx.build(targets='b')
        self.assertEqual(result, 130691232)

        # 'counter' node should not have been evaluated
        # and should still be stale
        self.assertEqual(C.counter, 0)

        # cannot use async solver here because C won't be
        # pickable from pyunit
        _ = afx.build()['counter']()
        self.assertEqual(C.counter, 1)

    def test_at_constructor(self):
        def subtract(p, q):
            return p - q

        afx = Artifax(
            ab=At('a', 'b', subtract),
            ba=At('b', 'a', subtract),
            a=-11,
            b=7.5,
        )

        ab, ba = afx.build(targets=('ab', 'ba'))
        self.assertEqual(ab, -18.5)
        self.assertEqual(ba, 18.5)



from inspect import getfullargspec
from functools import reduce
import os
from . import exceptions

_REPLACES = {
    '-': '_',
}

# make sure we can always undo a key replacement
assert len(_REPLACES) == len({v: k for k, v in _REPLACES.items()})

escape = lambda v: reduce(lambda s, k: s.replace(k, _REPLACES[k]), _REPLACES.keys(), v)
unescape = lambda v: reduce(lambda s, k: s.replace(_REPLACES[k], k), _REPLACES.keys(), v)

arglist = lambda v: (
    getfullargspec(v).args if callable(v) else
    v.args() if isinstance(v, At) else
    []
)


def to_graph(artifacts):
    """ returns a graph representation of the given artifacts """
    af_args = {k: arglist(v) for k, v in artifacts.items()}
    return {
        key: [k for k, v in af_args.items() if escape(key) in v]
        for key in artifacts
    }


def topological_sort(graph):
    """ returns a topological sorting of nodes from the given graph

    Throws artifax.CircularDependencyError
    if graph is not a Direct Acyclic Graph (DAG)
    """
    def _visit(node, temp, perm, tlist):
        if node in perm:
            return
        if node in temp:
            raise exceptions.CircularDependencyError('artifact graph is not a DAG')
        temp.add(node)
        for neighbor in graph[node]:
            _visit(neighbor, temp, perm, tlist)
        perm.add(node)
        tlist.insert(0, node)

    unmarked = set(graph.keys())
    temp = set()
    perm = set()
    tlist = []
    while unmarked:
        node = unmarked.pop()
        _visit(node, temp, perm, tlist)
    return tlist


def initial(graph):
    def to_mask(mask, node, graph, key):
        neighbors = graph[node]
        for neighbor in neighbors:
            mask[key[neighbor]] = 1
        return mask

    nodes = list(graph.keys())
    key = {nodes[i]: i for i in range(len(nodes))}
    mask = reduce(lambda m, n: to_mask(m, n, graph, key), nodes, [0] * len(nodes))

    return set([nodes[i] for i in range(len(mask)) if mask[i] == 0])


def pprint(*args, **kwargs):
    """ Prepends message with process id information """
    print('[{}]'.format(os.getpid()), end=' ')
    print(*args, **kwargs)


class At:
    """ The At class is but a wrapper for a tuple-like constructor
    that can be used to assign node values in a compact way without
    using lambda expressions.

    For example:
    {'a': lambda b, c: some_function(c, b)}

    can be rewritten as

    {'a': At('c', 'b', some_function)}
    """
    def __init__(self, *args):
        if len(args) < 2:
            raise ValueError('At constructor requires at least two arguments')
        self._args = args[:-1]
        self._value = args[-1]

    def args(self):
        """ returns list of lambda arguments """
        return self._args

    def value(self):
        """ returns lambda """
        return self._value
